# THIS FILE IS GENERATED FROM ORANGE-WIDGET-BASE SETUP.PY
short_version = '4.7.0'
version = '4.7.0'
full_version = '4.7.0'
git_revision = 'e7d65d993b276fc217262011098262eca739c936'
release = True

if not release:
    version = full_version
    short_version += ".dev"
